/*

Code for CSL model computation from paper

"An Efficient Parallel Algorithm for Multi-Scale Analysis of Connected Components in Gigapixel Images"

Michael H.F. Wilkinson, Martino Pesaresi, Georgios K. Ouzounis

IJGI (2016) in press.

Compile using

gcc -O3 -o parDAPSeg parDAPSeg.c -lm -lpthread

Run using

parDAPSeg <input image> <lambda file.lamvec> [imScale] [nthreads]

 */

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/times.h>
#include <unistd.h>
#include <math.h>
#include <assert.h>
//#include <xtiffio.h>



#define PI 3.14159265358979323846
#define false 0
#define true  1
#define MAXTHREADS 512
#define MAXLEVELS  256          /* altered to fit in 8 bit data */
#define CONNECTIVITY  6

typedef char bool;
typedef unsigned long ulong; 
typedef unsigned char ubyte;

typedef ubyte greyval_t;
typedef unsigned int area_t;
typedef unsigned int pixel_t;

#define MIN(a,b)  ((a<=b) ? (a) : (b))
#define MAX(a,b)  ((a>=b) ? (a) : (b))
#define LWB(self) (size2D*(((self)*depth)/nthreads))
#define UPB(self) (size2D*(((self+1)*depth)/nthreads))

#define bottom (0xffffffff)
#define ISROOT(a) ((a) == bottom) 
#define SETROOT(a) ((a) = bottom)
 
float imScale;
clock_t building_done, area_zones_done, build_and_filter_done, combine_images_start, combine_images_done;
clock_t start, MCCstart;
  
struct tms tstruct;

/************************** safe malloc and calloc ************************************/

//pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;

size_t total_alloc = 0, max_alloc=0;


void *Malloc(size_t n) {
  void *ptr;
  ptr = malloc(n);
  if (ptr==NULL) {
    fprintf (stderr, "Error: out of memory. "
	     "Could not allocate %lu bytes\n", n);
  }
  total_alloc += n;
  max_alloc = MAX(max_alloc,total_alloc);
  return ptr;
}

void *Calloc(size_t nmemb, size_t size) {
  void *ptr;
  ptr = calloc(nmemb, size);
  if (ptr==NULL) {
    fprintf (stderr, "Error: out of memory. Could not "
	     "allocate %lu bytes\n", nmemb*size);
  }
  total_alloc += nmemb*size;
  max_alloc = MAX(max_alloc,total_alloc);

  return ptr;
}

void Free( void *ptr, size_t size){
  free(ptr);
  total_alloc -= size;
}



typedef struct LambdaVec LambdaVec;

struct LambdaVec
{
   int NumLambdas;
   float *Lambda;    
}; 
/******************************************************************************/
/*                             Lambda Vector handling                         */
/******************************************************************************/

LambdaVec *LambdaVectorCreate(int size)
{
   LambdaVec *lvec;

   lvec = Malloc(sizeof(LambdaVec));
   if (lvec==NULL)  return(NULL);

   lvec->NumLambdas = size;
   lvec->Lambda = Malloc(size*sizeof(float));
   if (lvec->Lambda==NULL)
   {
     Free(lvec, sizeof(LambdaVec));
      return(NULL);
   }
   return(lvec);
}

void LambdaVectorDelete(LambdaVec *lvec)
{
  Free(lvec->Lambda, lvec->NumLambdas*sizeof(float));
  Free(lvec, sizeof(*lvec));
} 

LambdaVec *LambdaVectorRead(char *fname)
{
   FILE *infile;
   LambdaVec *l;
   int size, r;
   int c;

   infile = fopen(fname, "r");
   if (infile==NULL)  return(NULL);
   r = fscanf(infile, "LambdaVector\n");

   r = fscanf(infile, "%d\n", &size);

   l = LambdaVectorCreate(size);
   if(l)
   {

      for(c=0;c<size;c++)
	{   
	  r = fscanf(infile,"%f ",&l->Lambda[c]); 
	  l->Lambda[c] /= imScale;

	}  
   } 
   fclose(infile);
   return(l);
} 

int nthreads,numscales,offset=0;
pthread_t threadID[MAXTHREADS];

ubyte imgftype;
size_t width, height, depth, size;  /* precondition: width <= size/nthreads */
int size2D;
greyval_t *gval=NULL, *inverted=NULL,*out=NULL, *out2=NULL, *scales = NULL,
  *scales2 = NULL,*orig = NULL,*orig2 = NULL ;


area_t **lambda;
greyval_t sigma;
typedef struct MaxNode{ 
  pixel_t parent;
  area_t Area;
  greyval_t curDH;
  greyval_t scale;
  bool valid; 

} MaxNode;

MaxNode *node; 




/************************** semaphore ************************************/

pthread_mutex_t samut[MAXTHREADS];
pthread_cond_t  sacv[MAXTHREADS];
int             saval[MAXTHREADS];

void Psa(int p) {
  pthread_mutex_lock(&samut[p]);
  while (saval[p] <= 0)
    pthread_cond_wait(&sacv[p], &samut[p]);
  saval[p]--;
  pthread_mutex_unlock(&samut[p]);
}

void Vsa(int p) {
  pthread_mutex_lock(&samut[p]);
  saval[p]++;
  pthread_mutex_unlock(&samut[p]);
  pthread_cond_broadcast(&sacv[p]);
}

/************************** barrier ************************************/

pthread_mutex_t barriermutex = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t barriercv = PTHREAD_COND_INITIALIZER;
int barcnt = 0;

void Barrier(int self) {
  pthread_mutex_lock(&barriermutex);
  barcnt++;
  if (barcnt == nthreads) {
    barcnt = 0;  /* for reuse of routine */
    pthread_cond_broadcast (&barriercv);  
  } else {
    pthread_cond_wait (&barriercv, &barriermutex);
  }
  pthread_mutex_unlock(&barriermutex);
}

/************************** level root ************************************/

pixel_t levroot(pixel_t x) {
  pixel_t r=x, y ;
  greyval_t gv;
  if (ISROOT(r)) {
    return bottom;
  }
  gv = gval[x];
  while ((!ISROOT(node[r].parent)) && (gv==gval[node[r].parent]))
      r = node[r].parent;

  while (x!=r){
    y=node[x].parent;
    node[x].parent=r;
    x=y;
  }
  return r;
}

pixel_t Par(pixel_t x) {
  return levroot(node[x].parent);
}

void levrootfix(pixel_t lwb, pixel_t upb) {
  pixel_t z, u, x;

  for (x=lwb; x<upb;x++) {
    u = levroot(x);
    if (x!=u) node[x].parent=u;
    else {
      z = Par(x); 
      node[x].parent=z;
    }
  }
}


int findScale(area_t area, area_t *lambda, int numscales ){
  int upper = numscales-1, lower = 0, mid;
  
  //if (area <lambda[lower])
  //return lower;
  if (area >= lambda[upper])
    return upper+1;

  mid = (upper + lower) / 2;
  while(mid!=lower){
    if (area >= lambda[mid])
      lower = mid;
    else
      upper = mid;
    mid = (upper + lower) / 2;
  }
  return lower;
}

#define ISLEVELROOT(node, v)  ((ISROOT(node[v].parent)) || (gval[v]!=gval[node[v].parent])) 




void NodeSetDAPseg(uint lwb, uint upb, uint current, 
                   greyval_t *curDH,
		   greyval_t *maxDH, greyval_t *maxScale, 
		   greyval_t *maxOrig, greyval_t *curScale,  
		   greyval_t *outDH, greyval_t *outScale, 
		   greyval_t *outOrig, area_t *lambda, 
		   int numscales      ){
  /* pre: node[current].valid == false */
  pixel_t parent; 
  greyval_t scale, DH;

  if (ISLEVELROOT(node, current)){
    scale = findScale(node[current].Area,lambda,numscales);
    DH = ((!ISROOT(node[current].parent)) && (scale < numscales)) ?
      gval[current] - gval[node[current].parent] : 0;
  }
  if ( ISLEVELROOT(node, current) && 
       ( ( scale == numscales) || 
         ( ISROOT(node[current].parent) )       )  ){

    if ( !ISROOT(node[current].parent) ) {
      *maxScale = numscales;
      *maxOrig = 0;
    } else {
      *maxScale = scale;
      *maxOrig = gval[current];
    }

    *maxDH = 0;
    *curDH = 0;
    *curScale = *maxScale;

  } else {

    parent = node[current].parent;
    
    if (!node[parent].valid) { 
      /* go into recursion to set parent values correctly */

	NodeSetDAPseg( lwb, upb, parent, 
		       curDH, maxDH, maxScale, 
		       maxOrig, curScale,  
		       outDH, outScale, 
		       outOrig, lambda,numscales);

    } else { /* if the parent is valid, copy relevant values */

      *maxScale = outScale[parent];
      *maxDH = outDH[parent];
      *maxOrig = outOrig[parent];
      *curScale = node[parent].scale;
      *curDH = node[parent].curDH;

    } 
    
    if ( ISLEVELROOT(node, current) ){  
      /* if I have a level root, some things might change */
      
      if ( scale == *curScale) {
        /* parent's area is in same scale class, 
	   add current pixel's curDH */
	*curDH += DH;
      } else {
	/* at scale class change, update current scale and DH */
	*curDH = DH;
	*curScale = scale;
      }
      
      if ( *curDH >= *maxDH ) {
	/* If updated curDH is higher than or equal to the maximum DH found
	   update maxDH, maxScale, and outOrig */
	*maxDH = *curDH;
	*maxScale = *curScale;
	*maxOrig = gval[current];
      } 
      
    } 
 
  }

  if ( (lwb <= current) && (current < upb) ){
    /* store values if in current section */

    outScale[current] = *maxScale;
    outDH[current] = *maxDH;
    outOrig[current] = *maxOrig;      
    node[current].scale = *curScale;
    node[current].curDH = *curDH;
    node[current].valid = true;

  }
     
  return;
}


void MaxTreeSetDAPseg( int self,  
		       greyval_t *outDH, greyval_t *outScale, 
		       greyval_t *outOrig, area_t *lambda, 
		       int numscales     ){
  uint v, lwb, upb;
  greyval_t curScale, maxScale;
  greyval_t maxDH,  curDH, maxOrig;
  lwb = LWB(self); upb = UPB(self);
  for (v=lwb; v<upb; v++) {
    if (!node[v].valid) {
      NodeSetDAPseg(lwb, upb, v,
		    &curDH, &maxDH, 
		    &maxScale, &maxOrig, 
		    &curScale,  outDH, 
		    outScale, outOrig,lambda,numscales);
    }
  }
} /* MaxTreeSetDAPseg */






/************************** queue ************************************/

typedef struct Queue
{
  pixel_t *Pixels;
  size_t size;
  size_t Head;
  size_t Tail; /* First free place in queue, empty if Head=Tail */
} Queue;

Queue *QueueCreate(size_t imgsize)
{
   Queue *hq;

   hq = Calloc(MAXLEVELS, sizeof(Queue));
   if (hq==NULL) {
     fprintf (stderr, "Failure allocating hq. Size: %lu \n",
	      (size_t)MAXLEVELS * sizeof(Queue) );
 
       fprintf (stderr, "Out of memory!");
       return(NULL);
   }
   hq->Pixels = Calloc(imgsize, sizeof(pixel_t));
   if (hq->Pixels==NULL)
   {
     Free(hq, sizeof(Queue) );
      fprintf (stderr, "Failure allocating hq->Pixels. Size: %lu \n",
	      (size_t)imgsize * sizeof(uint) );
 
      fprintf (stderr, "Out of memory!");
      return(NULL);
   }
   hq->size = imgsize;
  return(hq);
} /* QueueCreate */

void FillPixelsInQueue(Queue *hq, pixel_t *numpixelsperlevel)
{
  int i;
  hq->Head = hq->Tail = 0;

  for (i=1; i<MAXLEVELS; i++)
    {
      hq[i].Pixels = hq[i-1].Pixels + numpixelsperlevel[i-1];
      hq[i].Head = hq[i].Tail = 0;
    }
} /* FillPixelsInQueue */

#define QueueFirst(hq,h)  (hq[h].Pixels[hq[h].Head++])
#define QueueAdd(hq,h,p)  hq[h].Pixels[hq[h].Tail++] = p
#define QueueNotEmpty(hq,h)  (hq[h].Head != hq[h].Tail)

void QueueDelete(Queue *hq)
{
  Free(hq->Pixels,hq->size*sizeof(pixel_t));

   Free(hq, sizeof(Queue));
} /* QueueDelete */

/****** Image read/write functions ******************************/

short ImagePGMAsciiRead(char *fname)
{
   FILE *infile;
   size_t i, r;
   int c;

   infile = fopen(fname, "r");
   if (infile==NULL) {
      fprintf (stderr, "Error: Can't read the ASCII file: %s !", fname);
      return(0);
   }
   r = fscanf(infile, "P2\n");
   while ((c=fgetc(infile)) == '#')
      while ((c=fgetc(infile)) != '\n');
   ungetc(c, infile);
   r = fscanf(infile, "%lu %lu\n255\n", &width, &depth);
   if ( r != 2 ) {
      fprintf (stderr, "Error reading file\n");
      fclose(infile);
      return(0);
   }   
   size = width*depth;
   height = 1;
   gval = Malloc(size*sizeof(greyval_t));
   if (gval==NULL) {
      fprintf (stderr, "Out of memory!\n");
      fclose(infile);
      return(0);
   }
   for (i=0; i<size; i++)
   {
      r = fscanf(infile, "%d", &c);
      gval[i] = c;
   }
   fclose(infile);
   return(1);
} /* ImagePGMAsciiRead */


short ImagePGMBinRead(char *fname)
{
   FILE *infile;
   int c, r;
   size_t i, readsize;
   ubyte *buf = NULL;

   infile = fopen(fname, "rb");
   if (infile==NULL) {
      fprintf (stderr, "Error: Can't read the binary file: %s !", fname);
      return(0);
   }
   r = fscanf(infile, "P5\n");
   while ((c=fgetc(infile)) == '#')
      while ((c=fgetc(infile)) != '\n');
   ungetc(c, infile);
   r = fscanf(infile, "%lu %lu\n255\n", &width, &depth);
   if ( r != 2 ) {
      fprintf (stderr, "Error reading file\n");
      fclose(infile);
      return(0);
   }   
   size = width*depth;
   height=1;
   buf = Malloc(size);
   if (buf==NULL) {
     fprintf (stderr, "Out of memory!");
     fclose(infile);
     return(0);
   } 
   gval = Malloc(size*sizeof(greyval_t));
   if (gval==NULL) {
     fprintf (stderr, "Out of memory!");
     fclose(infile);
     return(0);
   } 

   readsize = fread(buf, 1, size, infile);

   if (readsize!=size){
      fprintf (stderr, "Error reading file\n");
      fclose(infile);
      return(0);
   }   

   for (i=0; i<size; i++)
     gval[i]=buf[i];
   Free(buf, size);
   fclose(infile);
   return(1);
} /* ImagePGMBinRead */


short ImagePGMRead(char *fname)
{
   FILE *infile;
   char id[4];
   int r;

   infile = fopen(fname, "r");
   if (infile==NULL) {
      fprintf (stderr, "Error: Can't read the image: %s !", fname);
      return(0);
   }
   r = fscanf(infile, "%3s", id);
   if (r==0){
      fprintf (stderr, "Error reading file\n");
      fclose(infile);
      return(0);
   }   
   fclose(infile);
   if (strcmp(id, "P2")==0) return(ImagePGMAsciiRead(fname));
   else if (strcmp(id, "P5")==0) return(ImagePGMBinRead(fname));
   else {
     fprintf (stderr, "Unknown type of the image!");
     return(0);
   }
} /* ImagePGMRead */

int ImagePGMBinWrite(char *fname,  ubyte *buf)
{
   FILE *outfile;

   outfile = fopen(fname, "wb");
   if (outfile==NULL) {
      fprintf (stderr, "Error: Can't write the image: %s !", fname);
      return(-1);
   }
   fprintf(outfile, "P5\n%lu %lu\n255\n", width, depth);

   fwrite(buf, 1, (size_t)(size), outfile);
   fclose(outfile);
   return(0);
} /* ImagePGMBinWrite */


/****************** Construction of local tree using algorithm flood****************************/

int GetNeighbors(pixel_t p, pixel_t x, pixel_t y, pixel_t z, 
		 pixel_t *neighbors, pixel_t lwb, pixel_t upb)
{

   int n=0;

   x = p % width;

   if (x<(width-1))       neighbors[n++] = p+1;
   //if (y>0)               neighbors[n++] = p-width;
   if (x>0)               neighbors[n++] = p-1;
   //if (y<(height-1))      neighbors[n++] = p+width;
   if (depth>1) {
      if (z>lwb)          neighbors[n++] = p-size2D;
      if (z<(upb-1))           neighbors[n++] = p+size2D;
   }
   return(n);
} /* GetNeighbors */

int LocalTreeFlood(int self,  Queue *set, pixel_t *lero, int lev, area_t *thisarea)
{
   pixel_t neighbors[CONNECTIVITY];
   area_t area = *thisarea, childarea;
   pixel_t p, q, x, y=0, z, lwb, upb;
   int fq;
   int numneighbors, i, m;

   lwb = LWB(self);
   upb = UPB(self);

   while(QueueNotEmpty(set, lev)) {
      area++;
      p = QueueFirst(set, lev);
      x = p % width; 
      // y = (p % size2D) / width;
      z = p / size2D;
      numneighbors = GetNeighbors(p, x, y, z, neighbors, lwb/size2D, upb/size2D);
      for (i=0; i<numneighbors; i++) {
         q = neighbors[i];
         if (node[q].valid) {
            fq = gval[q];            
            node[q].valid = false;
            if (ISROOT(lero[fq])) lero[fq] = q;
            else node[q].parent = lero[fq];
            QueueAdd(set, fq, q);
            if (fq > lev) {
               childarea = 0;
               do {
                  fq = LocalTreeFlood(self, set, lero, fq, &childarea);
                  if (fq>=MAXLEVELS) {
                     return(fq);
                  }
               } while (fq!=lev);
               area += childarea;
            }
         }
      }
   }
   m = lev-1;
   while ((m>=0) && (ISROOT(lero[m])))  m--;
   if (m>=0) node[lero[lev]].parent = lero[m];
   node[lero[lev]].Area = area;
   lero[lev] = bottom;
   *thisarea = area;
   return(m);
} /* MaxTreeFlood */

/************************ Merge trees *************************************/

void Connect(pixel_t x, pixel_t y) {

  area_t area = 0, area1 = 0;
  pixel_t h, z;

  x = levroot(x);
  y = levroot(y);

  if (gval[x] < gval[y]) {
    h=x; x=y; y=h;
  }
  while ((x!=y) && (y!=bottom)) {
    z = Par(x);
    if ((z!=bottom) && (gval[z]>=gval[y])) {

      node[x].Area += area;
      x = z;

    } else {
      area1 = node[x].Area + area;
      area = node[x].Area ;
      node[x].Area =area1;
      node[x].parent = y;
      x = y;
      y = z;
    }
  }

  if (y==bottom) {
    while(x!=bottom) {
      node[x].Area += area;
      x = Par(x);
    }
  }
}


void Fuse(int self, int i) /* fuse regions [LWB(self), UPB(self+i-1)) and  [LWB(self+i), UPB(self+2i-1)) vertically */
{
  pixel_t p, q, x, y;

  greyval_t prevmin, curmin;

  /* get the horizontal boundary */ 

  p  = LWB(self+i);
  q = p - size2D;

  x = p % width;
  y = (p /width) % height;
  

  for ( y = 0 ; y < height ; y++ ){
    Connect(p, q);

    prevmin = MIN(gval[p],gval[q]);
    p++;
    q++;
    for ( x = 1 ; x < width ; x++, p++, q++ ){
      curmin = MIN(gval[p],gval[q]);
      
      if ((curmin > prevmin)){
	Connect(p, q);
      }
      prevmin = curmin;    
    }  



  }

  /*levrootfix( LWB(self), UPB(self+2*i-1) );*/
}


/****************** Concurrent construction and filter of Maxtree  ****************************/

typedef struct { 
  int self;
  Queue *thisqueue;
} ThreadData;

ThreadData *MakeThreadData(int numthreads){
  ThreadData *data = Malloc(numthreads *sizeof(ThreadData));
  int i;

  for (i=0; i<numthreads; i++){
    data[i].self=i;
    data[i].thisqueue= QueueCreate( UPB(i)-LWB(i));
    /* printf("Thread %d, LWB = %lu, UPB = %lu\n",i,LWB(i),UPB(i));*/
  }    
  return(data);
}

void FreeThreadData(ThreadData *data, int numthreads)
{
  int i;
  for (i=0;i<numthreads; i++)
    QueueDelete(data[i].thisqueue);
  Free(data, sizeof(ThreadData));
}

void combineResults(int self) {
  pixel_t p;
  pixel_t lwb =  LWB(self), upb = UPB(self);
  for (p = lwb; p<upb; p++){

    if (out2[p]>out[p]) { /* maxDH of opening > maxDH of closing instance */
      scales[p] = numscales + 1 + scales2[p];
      out[p] = out2[p];
      orig[p] = orig2[p];
    } else if (out2[p]<out[p]){
      orig[p] = MAXLEVELS - 1 - orig[p];
      scales[p] ++;            /* to match convention of Ispra code */
    } else { /* equality */
      
      scales[p]=0;
      orig[p] = (out[p]!=0) * (255-gval[p]);
    }
    
  }

}

void *ccaf(void *arg) {
  ThreadData *thdata = (ThreadData *) arg;
  int self = thdata->self, q, i;
  pixel_t x;
  area_t  area=0;
  pixel_t numpixelsperlevel[MAXLEVELS], lero[MAXLEVELS], xm;
  Queue *set = thdata->thisqueue;

  for (i=0; i<MAXLEVELS; i++) {
      numpixelsperlevel[i] = 0;
      lero[i] = bottom;
  }

  xm = LWB(self);

  for (x=xm; x<UPB(self); x++) { 
      numpixelsperlevel[gval[x]]++; 
      node[x].parent = bottom;
      node[x].valid = true;
      node[x].curDH = 0;
      if (gval[xm]>gval[x]) xm = x;
  }
  FillPixelsInQueue(set, numpixelsperlevel);
  
  QueueAdd(set, gval[xm], xm);
  node[xm].valid = false;
  lero[gval[xm]] = xm;             

  LocalTreeFlood(self, set, lero, gval[xm], &area);


  /*  for (x=LWB(self); x<UPB(self); x++) { 
      node[x].valid = false;
      }*/
  i = 1;
  q = self;
  while ((self+i<nthreads) && (q%2 == 0)) {
    Psa(self+i);  /* wait to glue with righthand neighbor */
    Fuse(self, i); 
    i = 2*i;
    q = q/2;
  }
  if (self != 0) {
    Vsa(self);  /* signal lefthand neighbor */
  }

  Barrier(self); 
  if (self == 0){
    area_zones_done = building_done = times(&tstruct);
  }

  /*MaxTreeAreaZones(self, lambda[self], numscales);
    Barrier(self); */

  MaxTreeSetDAPseg( self, out, scales, orig, lambda[self], numscales);

  if (gval == inverted)
    combineResults(self);
  
  return NULL;
}


void BuildTreeAndFilter(ThreadData *thdata, int nthreads) {
  int thread;
  for (thread=0; thread<nthreads; ++thread) {
    pthread_create(threadID+thread, NULL, ccaf, (void *) (thdata + thread));
  }
  for (thread=0; thread<nthreads; ++thread) {
    pthread_join(threadID[thread], NULL);
  }
}



int main (int argc, char *argv[]) {
  
  char *imgfname, *lambdafname;
  greyval_t *temp;
  ThreadData *thdata;
  LambdaVec *L;
  pixel_t i,j;
  long tickspersec = sysconf(_SC_CLK_TCK);  
  float musec;
  float buildingtime, areazonetime, MOCsettime, combineimagestime;
  sigma =0;
  imScale = 0.5;
  if (argc<=2)
    {
      printf("Usage: %s <input image> <lambda file.lamvec> [imScale] [nthreads]\n", argv[0]);
      exit(0);
    }
  
  imgfname                    = argv[1];
  if (argc>2)  lambdafname   = argv[2];
  if (argc>3)
    imScale = atof(argv[3]);  
 
  if (imScale==0) imScale = 0.5;

  imScale*= imScale;

  if (argc>4){
    nthreads = MIN(atoi(argv[4]), MAXTHREADS);
  } else {
    if (getenv("NCPUS")!=NULL){
      nthreads = MIN(atoi(getenv("NCPUS")), MAXTHREADS);
    } else {nthreads = 1;}
  }
 
  fprintf(stderr, "nthreads = %d\n",nthreads); 
  
  
  /* Load Configuration File */
  L = LambdaVectorRead(lambdafname);
  numscales = L->NumLambdas-1;
  lambda = malloc(nthreads * sizeof(area_t*));

  for (i = 0; i<nthreads; i++){
    lambda[i] = malloc(numscales*sizeof(area_t));
    for (j=0;j<numscales;j++)
      lambda[i][j] = (area_t)floor(L->Lambda[j+1]);
  }

  fprintf(stderr, "reading image %s\n",imgfname); 
  if (!ImagePGMRead(imgfname)) return(-1);
  size2D = width;
   fprintf(stderr, "reading image %s done\n",imgfname); 
 
  
   fprintf(stderr, "Image: Width=%lu Height=%lu Depth=%lu\n", width, height, depth);
  printf ("%d, ", nthreads);
  inverted = Calloc((size_t)size, sizeof(greyval_t));
  for (i = 0; i<size;i++)
    inverted[i]= 255-gval[i];
  
  node = Calloc((size_t)size, sizeof(MaxNode));
  if (node==NULL) {
    fprintf(stderr,"Tried to allocate = %ld\n",size * sizeof(MaxNode) );    
    fprintf(stderr, "out of memory! \n");
    free(gval);
    return(-1);
  }

 
  out =  Malloc(size*sizeof(greyval_t));
  if (out==NULL) {
    fprintf(stderr,"Tried to allocate = %ld\n",size * sizeof(greyval_t) );    
    fprintf(stderr, "Can't create out! \n");

    free(node);
    free(gval);
    return(-1);
  }

  out2 =  Malloc(size*sizeof(greyval_t));
  if (out2==NULL) {
    fprintf(stderr,"Tried to allocate = %ld\n",size * sizeof(greyval_t) );    
    fprintf(stderr, "Can't create out2! \n");

    free(node);
    free(out);
    free(gval);
    return(-1);
  }
  scales =  Malloc(size*sizeof(greyval_t));
  if (scales==NULL) {
    fprintf(stderr,"Tried to allocate = %ld\n",size * sizeof(greyval_t) );    
    fprintf(stderr, "Can't create scales! \n");

    free(node);
    free(gval);
    free(out);
    free(out2);
    return(-1);
  }
  scales2 =  Malloc(size*sizeof(greyval_t));
  if (scales2==NULL) {
    fprintf(stderr,"Tried to allocate = %ld\n",size * sizeof(greyval_t) );    
    fprintf(stderr, "Can't create scales2! \n");
    free(node);
    free(gval);
    free(out);
    free(out2);
    free(scales);
    return(-1);
  }
  
  orig =  Malloc(size*sizeof(greyval_t));
  if (orig ==NULL) {
    fprintf(stderr,"Tried to allocate = %ld\n",size * sizeof(greyval_t) );    
    fprintf(stderr, "Can't create orig! \n");

    free(node);
    free(gval);
    free(out);
    free(out2);
    free(scales);
    free(scales2);
    return(-1);
  }

  orig2 =  Malloc(size*sizeof(greyval_t));
  if (orig2==NULL) {
    fprintf(stderr,"Tried to allocate = %ld\n",size * sizeof(greyval_t) );    
    fprintf(stderr, "Can't create orig2! \n");

    free(node);
    free(gval);
    free(out);
    free(out2);
    free(scales);
    free(scales2);
    free(orig);
    return(-1);
  }
  
  thdata = MakeThreadData(nthreads); 
  fprintf(stderr, "Data read, start filtering.\n");
  start = times(&tstruct);
  
  BuildTreeAndFilter(thdata,nthreads);
  fprintf(stderr, "Max-Tree built.\n");

  build_and_filter_done = times(&tstruct);
  buildingtime = (float)(building_done - start)/((float)tickspersec);
  areazonetime = (float)(area_zones_done - building_done)/((float)tickspersec);
  MOCsettime = (float)(build_and_filter_done - area_zones_done)/((float)tickspersec);

  FreeThreadData(thdata,nthreads);

  free(gval);
  
  gval =inverted;
  offset=numscales;
  
  temp = out;
  out = out2;
  out2 = temp;
  temp =scales;
  scales = scales2;
  scales2 = temp;
  temp =orig;
  orig = orig2;
  orig2 = temp;
  
  /* memset(node, 0, (size_t)size * sizeof(MaxNode));*/


 
 
   thdata = MakeThreadData(nthreads); 
   fprintf(stderr,"Data read, start filtering.\n");

   MCCstart = times(&tstruct);
   
   BuildTreeAndFilter(thdata,nthreads);
  fprintf(stderr, "Min-Tree built.\n");

   build_and_filter_done = times(&tstruct);
   buildingtime += (float)(building_done - MCCstart)/((float)tickspersec);
   areazonetime += (float)(area_zones_done - building_done)/((float)tickspersec);
   MOCsettime += (float)(build_and_filter_done - area_zones_done)/((float)tickspersec);

   combine_images_start = times(&tstruct);
   //CombineImages(thdata,nthreads);
  fprintf(stderr, "Images combined.\n");
   combine_images_done = times(&tstruct);
   
   combineimagestime = (float)(combine_images_done - combine_images_start)/((float)tickspersec);
   
   musec = (float)(times(&tstruct) - start)/((float)tickspersec);
   
   ImagePGMBinWrite("C.pgm",out);
   ImagePGMBinWrite("L.pgm",orig);
   ImagePGMBinWrite("S.pgm",scales);
   
   FreeThreadData(thdata,nthreads);

   free(node);
   free(gval);


  for (i = 0; i<nthreads; i++){
    free(lambda[i]);
  }
   free(lambda);
 
   fprintf(stderr, "total allocation: %lu bytes = %f MB\n",max_alloc,(float)max_alloc/(1024.0*1024.0) );

   printf(" %f, ",musec);
   printf(" %f, ",buildingtime);
   printf(" %f, ",areazonetime);
   printf(" %f, ",MOCsettime);
   printf(" %f \n",combineimagestime);
  return(0);
} /* main */
